#!/usr/bin/env python3
import json
import multiprocessing
import os
import signal
import subprocess
import logging
import time
from time import sleep
import netifaces as net
from ipaddress import ip_address, IPv6Address

import requests
from configobj import ConfigObj
from sysrepo_api import sysrepo_api3 as api
from sysrepo_systemresources import sysrepo_systemresources as systemresources_api

default_k8s_controller = "telco-systems-k8s-controller"
ucpe_cfg_file="/config/ucpe_config.json"
K8S_DISABLED_FILE='/config/k8s-disabled'
K8S_CENTRAL_BOX="/config/central-box"
KUBELET_CONFIG = "/etc/systemd/system/kubelet.service.d/10-kubeadm.conf"
K8S_SINGLE_MACHINE_FILE = '/config/k8s-single-machine'
CALICO_VPP_CONFIG = '/kubernetes/templates/calico-vpp.yaml'
BIND_INTERFACE_FILE = '/kubernetes/bind-interface-file'
DHCP_CONFIGS = '/kubernetes/dhcp'

def cluster_join_process(self, controller):
    join_count = 0

    while True:
        join_count += 1
        try:
            self.logger.debug("uCPE is going to join controller {}, join retry {}".format(controller, join_count))
            token = self.get_auth_token_tveapi_controller(controller)  # get auth token
            if token is None:
                raise Exception("Could not get tveapi token!")

            data = self.get_auth_token_kubernetes(token, controller)  # get auth token
            if data is None:
                raise Exception("Could not get kubernetes token!")
            else:
                token_k8s, cert = data
                self.logger.debug("kubernetes cluster token: %s, cert: %s" % (token_k8s, cert))

            net_mode = subprocess.check_output(['sysrepocfg', '-X', '-x', '/ucpe:config//system:system//network-mode', '-f', 'json', '-d', 'startup']).decode("utf-8")
            self.logger.debug("net_mode: {} ".format(net_mode))
            # if net_mode.find('"network-mode": "vpp"') > -1:
            #     self.set_main_vpp_int(token, controller)

            dns = os.popen("resolvectl status | grep br-data0 -A 3 | awk 'NR==4{print $4}'").read().rstrip('\n')
            self.logger.debug("uCPE join service - dns: {}".format(dns))

            #subprocess.check_call(['kubeadm', 'join', '{}:6443'.format(controller), '--token', '{}'.format(token_k8s),
            # '--discovery-token-unsafe-skip-ca-verification', '> /dev/null', '2>&1'])
            self.logger.debug("uCPE join cluster retry number %d" % join_count)
            self.execute_shell_command("kubeadm join {}:6443 --token {} --discovery-token-ca-cert-hash {} > /dev/null 2>&1".format(controller, token_k8s, cert))
            system = systemresources_api.SysrepoSystemResources(app_name="set systemresources")
            system.refresh_k8s_resources()
            #res = os.system("ifconfig | grep vpp1 >/dev/null 2>&1")
            #count = 15
            #while count >= 0:
            #    self.logger.debug("uCPE vppctl stn rule retry %d" % count)
            #    if not res:
            #        ip = os.popen('ifconfig vpp1 | grep "inet"').read().split()[1]
            #        if ip:
            #            if dns:
            #                os.system('systemd-resolve --interface vpp1 --set-dns {}'.format(dns))
            #                self.logger.debug("uCPE join service - dns configured: {}".format(dns))
            #            os.system('vppctl stn rule address {} interface tap0'.format(ip))
            #            break
            #    count -= 1
            #    sleep(10)
            #    res = os.system("ifconfig | grep vpp1 >/dev/null 2>&1")

            break
        except Exception as ex:
            self.logger.error('cluster join failed with error: {}'.format(ex))
            wait_sec = 40
            self.logger.error("Waiting {} seconds for the next try to join kubernetes cluster ...".format(wait_sec))
            time.sleep(wait_sec)

class SysrepoKubernetes(api.SysrepoAPI):
    threads_flag = None

    def __init__(self,
                 app_name="kubernetes",
                 log_file=None,
                 #log_level=None,
                 log_level=logging.DEBUG,
                 console_log=False,
                 namespace="/ucpe:config/kubernetes:kubernetes",
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            namespace=namespace,
            run_as_subscriber=run_as_subscriber,
            priority=priority)
        self.join_process=None

    def apply_startup(self):
        self.logger.debug('Apply STARTUP config for Kubernetes')
        json_data = self.get_startup_config(full_config=False, full_path=False)
        cbox_enabled = False
        controller = default_k8s_controller

        if os.path.exists(K8S_CENTRAL_BOX):
            pass
        elif os.path.exists(K8S_DISABLED_FILE):
            self.logger.info('System capabilities kubernetes disabled')
            return True

        self.logger.debug('Startup config : %s' % json.dumps(json_data))
        if 'cbox' in json_data:
            if 'enabled' in json_data['cbox']:
                cbox_enabled = json_data['cbox']['enabled']
        if 'controller' in json_data:
            controller = json_data['controller']

        if not os.path.isfile(BIND_INTERFACE_FILE) or os.stat(BIND_INTERFACE_FILE).st_size == 0:
            self.__bind_interface(json_data['bind-interface'])

        # TODO: ippass file should exist prior cluster creation
        self.__interfaces_check_ippass()

        if cbox_enabled:
            # CBOX
            self.logger.debug('Update CBOX state \'{}\' for Kubernetes '.format(cbox_enabled))
            if not os.path.exists(K8S_CENTRAL_BOX):
                if not os.path.exists(K8S_SINGLE_MACHINE_FILE):
                    with open(K8S_SINGLE_MACHINE_FILE, 'a'):
                        os.utime(K8S_SINGLE_MACHINE_FILE, None)
        else:
            # Remote Controller
            self.logger.debug('Update non CBOX state \'{}\' for Kubernetes '.format(cbox_enabled))

        if os.path.exists(K8S_CENTRAL_BOX) or os.path.exists(K8S_SINGLE_MACHINE_FILE):
            self.set_collectd_controller_mode(is_controller = True)

        for dhcp_config in os.listdir(DHCP_CONFIGS):
            self.logger.info(f'Starting k8s DHCP server on interface {dhcp_config}')
            self.execute_shell_command(
                f'python3 /scripts/kubernetes_dhcp.py {DHCP_CONFIGS} {dhcp_config} IN_CREATE > /dev/null 2>&1')
            self.execute_shell_command(
                f'python3 /scripts/kubernetes_dhcp.py {DHCP_CONFIGS} {dhcp_config} IN_CLOSE_WRITE > /dev/null 2>&1')

        controller = self.get_formatted_ip_address(controller)

        if os.path.exists(K8S_CENTRAL_BOX) or os.path.exists(K8S_SINGLE_MACHINE_FILE):
            controller = self.get_item("/ucpe:config/system:system/hostname")
            self.create_registry(controller)
            return True

        is_joined = False
        try:
            state = subprocess.check_output(['systemctl', 'is-active', 'kubelet.service'])
            if state.decode().rstrip() == 'active':
                if os.path.exists('/etc/kubernetes/pki/ca.crt'):
                    is_joined = True
        except :
            pass

        if not is_joined:
            # if there is ongoing join process let it finish!
            if self.join_process is not None and self.join_process.pid is not None:
                self.logger.debug('Apply startup config - there is already running join process!')
                #self.join_process.terminate()
                return True

            self.logger.debug('Apply startup config - starting new join process')
            self.start_cluster_join_service(controller)
        else:
            self.logger.debug('Apply startup config - worker already joined to %s' % ( controller ))

        return True

    def apply_running(self, change_set):
        self.logger.debug('Apply RUNNING config for Kubernetes')
        json_data = self.get_running_config(full_path=False)
        cbox_enabled = False
        controller = default_k8s_controller

        if os.path.exists(K8S_CENTRAL_BOX):
            pass
        elif os.path.exists(K8S_DISABLED_FILE):
            self.logger.info('System capabilities kubernetes disabled')
            return True

        if 'cbox' in json_data:
            if 'enabled' in json_data['cbox']:
                cbox_enabled = json_data['cbox']['enabled']
        if 'controller' in json_data:
            controller = json_data['controller']

        for change in change_set:
            self.logger.debug('Operation -->> %s' % change['operation'])
            self.logger.debug('XPath -->> %s' % change['xpath'])
            if change['operation'] == 'MODIFIED':
                if "bind-interface" in change["xpath"]:
                    self.logger.debug(
                        'Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    if self.join_process is not None and self.join_process.pid is not None:
                        self.logger.debug('apply runnnig config - terminate running join')
                        self.join_process.terminate()

                    self.cluster_reset()  # TODO: WAIT AFTER AGENT CLEANUP TO AVOID RESCHEDULING
                    self.__bind_interface(change['new_value'])
                    if os.path.isfile(K8S_SINGLE_MACHINE_FILE):
                        self.restart_nfvtime_kubernetes_init()
                        # workaround due to containerd failing - reboot required
                        self._system_reboot()
                    else:
                        self.clean_worker_in_kube_cluster(self.get_controller())
                        if os.path.exists(K8S_CENTRAL_BOX):
                            self.execute_shell_command('systemctl enable nfvtime-kubernetes-init')

                        self._system_reboot()

                if "cbox/enabled" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change[ 'xpath' ], change[ 'old_value' ], change[ 'new_value' ]) )
                    if 'true' in str(change['new_value']).lower():
                        cbox_enabled = True
                        if not os.path.exists(K8S_CENTRAL_BOX):
                            with open(K8S_SINGLE_MACHINE_FILE, 'a'):
                                os.utime(K8S_SINGLE_MACHINE_FILE, None)

                        self.execute_shell_command('systemctl enable nfvtime-kubernetes-init')
                        if os.path.exists(K8S_CENTRAL_BOX):
                            self.execute_shell_command('kubectl taint nodes --all node-role.kubernetes.io/master-')

                        self.set_collectd_controller_mode(is_controller = True)
                        try:
                            if os.stat('/kubernetes/calico-ready').st_size != 0:
                                self.cluster_reset()
                        except Exception:
                            pass
                        self._system_reboot()

                    else:
                        cbox_enabled = False
                        if os.path.exists(K8S_SINGLE_MACHINE_FILE):
                            self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                            os.remove(K8S_SINGLE_MACHINE_FILE)

                        self.set_collectd_controller_mode(is_controller = False)
                        self.cluster_reset()
                        self._system_reboot()

                    self.change_hp_count()

                if os.path.exists(K8S_CENTRAL_BOX) or os.path.exists(K8S_SINGLE_MACHINE_FILE):
                    continue

                if "controller" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change[ 'xpath' ], change[ 'old_value' ], change[ 'new_value' ]) )
                    if os.path.exists(K8S_SINGLE_MACHINE_FILE):
                        self.logger.info('CBOX mode - Controller value will take effect when CBOX mode in disabled.')
                        continue

                    old_controller = change[ 'old_value' ]
                    old_controller = self.get_formatted_ip_address(old_controller)
                    controller = change[ 'new_value' ]

                    # if there is ongoing join process terminate it!
                    if self.join_process is not None and self.join_process.pid is not None:
                        self.logger.debug('apply runnnig config - terminate running join')
                        self.join_process.terminate()

                    self.cluster_reset()
                    self.clean_worker_in_kube_cluster(old_controller)
                    #net_mode = subprocess.check_output(['sysrepocfg', '-g', '/ucpe:config//system:system//network-mode', '-f', 'json', '-d', 'startup']).decode("utf-8")
                    #if net_mode.find('"network-mode": "vpp"') > -1:
                    #    token = self.get_auth_token_tveapi_controller(controller)  # get auth token
                    #    if token:
                    #        self.set_main_vpp_int(token, controller, 0)

                    #self.set_controller(controller)
                    #self.cluster_join()
                    #self.logger.debug('apply runnnig config - start new join')
                    self.logger.debug('apply runnnig config - REBOOTING before starting new join')
                    self.logger.info('Calling system_reboot !')
                    self._system_reboot()
                    #self.start_cluster_join_service(controller)

            if change['operation'] == 'CREATED':
                if "controller" in change["xpath"]:
                    self.logger.debug('Created  - %s: %s' % (change[ 'xpath' ], change[ 'new_value' ]) )
                    if os.path.exists(K8S_SINGLE_MACHINE_FILE):
                        self.logger.info('CBOX mode - Controller value will take effect when CBOX mode in disabled.')
                        continue

                    controller = change[ 'new_value' ]

                    # if there is ongoing join process terminate it!
                    if self.join_process is not None and self.join_process.pid is not None:
                        self.logger.debug('apply runnnig config - terminate running join')
                        self.join_process.terminate()

                    self.cluster_reset()
                    #net_mode = subprocess.check_output(['sysrepocfg', '-g', '/ucpe:config//system:system//network-mode', '-f', 'json', '-d', 'startup']).decode("utf-8")
                    #if net_mode.find('"network-mode": "vpp"') > -1:
                    #    token = self.get_auth_token_tveapi_controller(controller)  # get auth token
                    #    if token:
                    #        self.set_main_vpp_int(token, controller, 0)

                    #self.set_controller(controller)
                    #self.cluster_join()
                    #self.logger.debug('apply runnnig config - start new join')
                    self.logger.debug('apply runnnig config - REBOOTING before starting new join')
                    self.logger.info('Calling system_reboot !')
                    self._system_reboot()
                    #self.start_cluster_join_service(controller)

            if change['operation'] == 'DELETED':
                if "controller" in change["xpath"]:
                    self.logger.debug('Deleted  - %s' % change[ 'xpath' ] )

                    startup_config = self.get_startup_config(full_config=False, full_path=False)

                    if 'controller' in startup_config:
                        old_controller = startup_config['controller']
                        self.logger.debug('Deleted worker from controller  - %s' % old_controller )
                        old_controller = self.get_formatted_ip_address(old_controller)

                        self.clean_worker_in_kube_cluster(old_controller)

                    # vstojkov: NFV-19128 - call cluster_reset in case: central to cbox device
                    # if os.path.exists(K8S_SINGLE_MACHINE_FILE):
                    #     self.logger.info('CBOX mode - Controller value will take effect when CBOX mode in disabled.')
                    #     continue

                    # if there is ongoing join process terminate it!
                    if self.join_process is not None and self.join_process.pid is not None:
                        self.logger.debug('apply runnnig config - terminate running join')
                        self.join_process.terminate()
                    try:
                        if os.stat('/kubernetes/calico-ready').st_size != 0:
                            self.cluster_reset()
                    except Exception:
                        pass
                    self.logger.info('Calling system_reboot !')
                    self._system_reboot()

        return True

    def apply_config(self, change_set=None, apply_startup=False):
        ret = False

        if apply_startup:
            ret = self.apply_startup()
        else:
            ret = self.apply_running(change_set)
        return ret

    def set_bind_interface(self, value="data0"):
        self.set_item("{}/bind-interface".format(self.namespace), value)

    def set_controller(self, value='telco-systems-k8s-controller'):
        self.set_item("{}/controller".format(self.namespace), value)

    def get_controller(self):
        return self.get_item("{}/controller".format(self.namespace))

    def delete_controller(self):
        self.delete_item("{}/controller".format(self.namespace))

    def set_cbox_enabled(self, value=False):
        self.set_item("{}/cbox/enabled".format(self.namespace), value)

    def get_cbox_enabled(self):
        return self.get_item("{}/cbox/enabled".format(self.namespace))

    def get_worker_name(self):
        return self.get_item("/ucpe:config/system:system/hostname")

    def get_interfaces(self):
        self.namespace = "/ucpe:config/interfaces:interface"
        config_indexed = {}
        for interface in self.get_running_config(full_path=False):
            config_indexed[interface['name']] = interface

        self.namespace = "/ucpe:config/kubernetes:kubernetes"
        return config_indexed


    def get_memory_profile(self):
        mem_total = 0
        hp_count = 0
        hp_size = 0
        with open('/proc/meminfo') as fl:
            for line in fl:
                if 'MemTotal' in line:
                    mem_total = int(line.split()[1])
                if 'HugePages_Total' in line:
                    hp_count = int(line.split()[1])
                if 'Hugepagesize' in line:
                    hp_size = int(line.split()[1])

            fl.close()
        return (mem_total, hp_count, hp_size)


    def change_hp_count(self):
        new_hp_count = 512
        if os.path.isfile("/config/k8s-single-machine"):
            new_hp_count = 2048

        (mem_total, hp_count, hp_size) = self.get_memory_profile()

        if hp_count >= new_hp_count:
            return

        self.logger.info("Requesting huge-page-count %s" % str(new_hp_count))
        try:
            os.mknod("/config/_reboot/sysres/%s" % str(new_hp_count))
        except:
            self.logger.info(f"Error creating file \"/config/_reboot/sysres/{str(new_hp_count)}\"")

    def set_collectd_controller_mode(self, is_controller):
        if os.path.exists('/etc/collectd/collectd.conf'):
            if is_controller == True:
                # Enable k8s_stats plugin for controller and cbox
                self.execute_shell_command('sed -i "/Import \\"k8s_stats_plugin\\"/c\  Import \\"k8s_stats_plugin\\"" /etc/collectd/collectd.conf')
                self.execute_shell_command('sed -i "/Import \\"k8s_notify_plugin\\"/c\  Import \\"k8s_notify_plugin\\"" /etc/collectd/collectd.conf')
            else:
                self.execute_shell_command('sed -i "/Import \\"k8s_stats_plugin\\"/c\# Import \\"k8s_stats_plugin\\"" /etc/collectd/collectd.conf')
                self.execute_shell_command('sed -i "/Import \\"k8s_notify_plugin\\"/c\# Import \\"k8s_notify_plugin\\"" /etc/collectd/collectd.conf')

            self.execute_shell_command('systemctl --no-block restart collectd 2>/dev/null || true')

    def get_auth_token_tveapi_controller(self, controller):
        cfg_obj = ConfigObjNet("/root/local-openrc.env")
        if cfg_obj and cfg_obj['OS_USERNAME'] and cfg_obj['OS_USERNAME'] != "":
            user = cfg_obj['OS_USERNAME']
        else:
            user = "admin"
        if cfg_obj and cfg_obj['OS_PASSWORD'] and cfg_obj['OS_PASSWORD'] != "":
            password = cfg_obj['OS_PASSWORD']
        else:
            password = "admin"

        #controller = self.get_controller()
        url = "https://%s:17374/v3/auth/tokens" % controller
        headers = {'Content-type': 'application/json', }
        data = '{"user":"%s","password":"%s"}' % ( user, password )

        for attemt in range(2):
            try:
                response = requests.get(url, headers=headers, data=data, verify=False, auth=('admin', password),timeout=10)
                self.logger.debug("get_auth_token_tveapi_controller response: %s" % response.text)
                #self.logger.debug("get_auth_token_tveapi_controller response (token): %s" % response.headers['X-Subject-Token'])
                return response.text
            except requests.ConnectionError:
                self.logger.error("Connection to \"%s\" FAILED!" % url)
                sleep(4)
            except requests.Timeout:
                self.logger.error("Connection to \"%s\" TIMED OUT!" % url)
                sleep(4)
            except:
                self.logger.error("Failed to get authentication token!")
                sleep(4)
        else:
            self.logger.error("Failed to get authentication token 2 attempts!")

    def get_auth_token_kubernetes(self, token, controller):
        if token is None:
            return

        #controller = self.get_controller()
        url = "https://%s:7373/v1/kubernetes/token" % controller
        headers = {'Content-type': 'application/json', 'X-Auth-Token': token, }
        params = ()

        try:  # get kubernetes token for uCPE join
            response = requests.get(url, headers=headers, params=params, verify=False, timeout=10)
            token = response.text.split()[0]
            cert = response.text.split()[1]
            self.logger.debug("get_auth_token_kubernetes response: %s" % response.text)
            self.logger.debug("get_auth_token_kubernetes token: %s cert: %s" % (token, cert))
            data = [token, cert]
            return data

        except:
            self.logger.error("Error occurred while getting token from kubernetes controller.")

    def set_main_vpp_int(self, token, controller, cmd=1):
        if token is None:
            return

        name = self.get_worker_name()
        #controller = self.get_controller()
        url = "https://%s:7373/v1/kubernetes/main_vpp_int" % controller
        headers = {'Content-type': 'application/json', 'X-Auth-Token': token, }

        if os.path.exists('/config/customize/kubernetes/port'):
            with open('/config/customize/kubernetes/port') as f:
                pci_id = f.read()
        else:
            pci_id = subprocess.check_output("lshw -class network -businfo | grep pci | tail -1 | awk '{print $1}' | cut -d@ -f 2 | tr -d '\n'", shell=True).decode()
        if pci_id is None:
            return
        len_id = len(pci_id)
        pci_id = pci_id[:(len_id-1)] + '*' + pci_id[(len_id-1):]
        show_cmd="/usr/bin/vppctl show hardware-interface | grep -B8 %s | awk 'NR==1 {{print $1}}' | tr -d '\\n'" % pci_id
        interface = self.execute_shell_command(show_cmd)

        if cmd == 1:
            data = '{"name":"%s", "int_name":"%s", "cmd":"add"}' % (name, interface)
        else:
            data = '{"name":"%s", "int_name":"%s", "cmd":"remove"}' % (name, interface)

        try:
            response = requests.put(url, headers=headers, data=data, verify=False, timeout=10)
            self.logger.debug("set_main_vpp_int response: %s" % response.text)
            if response.status_code == requests.codes.ok:
                self.logger.debug("set_main_vpp_int success!")
            return response.text
        except:
            self.logger.error("Error occurred while set_main_vpp_int().")

    def create_registry(self, controller='telco-systems-k8s-controller'):
        insecure_registries=[f'{controller}:5001']
        # Get controller IP address
        if os.path.isfile(BIND_INTERFACE_FILE) and os.stat(BIND_INTERFACE_FILE).st_size != 0:
            with open(BIND_INTERFACE_FILE) as f:
                bind_interface = f.read()
            try:
                for version in 4,6:
                    ip = os.popen(f'ip -{version} addr show dev {bind_interface} scope global | grep inet').read().split()[1]
                    if ip != []:
                        ip = self.get_formatted_ip_address(ip.split('/')[0])
                        insecure_registries.append(f'{ip}:5001')
            except IndexError:
                pass

        for registry in insecure_registries:
            dir = os.path.join('/etc/containerd/certs.d/' + registry)
            if not os.path.exists(dir):
                os.makedirs(dir, exist_ok=True)
                f = open(os.path.join(dir + '/hosts.toml'), 'w')
                f.write(f'server = "http://{registry}"\n\n[host."http://{registry}"]\n  insecure_skip_verify = true')
                f.close()

    def get_worker_status_from_kubernetes(self, token=None):
        if token is None:
            return

        name = self.get_worker_name()
        controller = self.get_controller()
        self.logger.debug('Worker {} going to send request for status to controller {}'.format(name,controller))
        url = "https://%s:7373/v1/kubernetes/worker_status" % controller
        headers = {'Content-type': 'application/json', 'X-Auth-Token': token, }
        params = (('name', name),)

        try:
            response = requests.get(url, headers=headers, params=params, verify=False, timeout=10)
            self.logger.debug("get_worker_status response: %s" % response.text)
            #self.logger.error("get_worker_status response: %s" % response.headers['X-Subject-Token'])
            return response.text
        except:
            self.logger.error("Error occurred while getting worker status from kubernetes controller.")

    def clean_worker_in_kube_cluster(self, controller, workers=''):
        self.logger.debug('Worker going to send request for clean kubernetes controller: {}'.format(controller))
        token = self.get_auth_token_tveapi_controller(controller) # get auth token
        if not token:
            self.logger.error('Can not get token from {}'.format(controller))
            return False

        name = self.get_worker_name()
        workers = [name] if not workers else workers.split(',')

        url = "https://%s:7373/v1/kubernetes/worker_clean" % controller
        headers = {'Content-type': 'application/json', 'X-Auth-Token': token, }
        data = json.dumps(workers)
        result = ''

        try:
            response = requests.put(url, headers=headers, data=data, verify=False, timeout=10)
            self.logger.debug("clean_worker_in_kube_cluster response: %s" % response.text)
            result = response.text
            if response.status_code == requests.codes.ok:
                self.logger.debug("clean_worker_in_kube_cluster success!")
            else:
                self.logger.error("fail worker_clean")
                return False, result

            return True, result
        except:
            self.logger.error("Error occurred while clean_worker_in_kube_cluster().")
            return False, result

    def cluster_reset(self): # TODO: WAIT AFTER AGENT CLEANUP TO AVOID RESCHEDULING
        self.logger.debug("Inside worker cluster_reset()")
        pci_addr = self.execute_shell_command("dpdk-devbind --status-dev net | grep drv=vfio-pci | awk '{ ORS=\" \" }; {print $1}'")
        try:
            self.logger.info("uCPE is going to cluster reset")
            # Send SIGTERM to calico-vpp-agent for graceful stop
            agent_pid = self.execute_shell_command("pidof calico-vpp-agent")
            if agent_pid:
                self.logger.debug("Killing calico-agent with PID {}".format(agent_pid))
                os.kill(int(agent_pid), signal.SIGTERM)

            self.execute_shell_command("truncate -s 0 /kubernetes/calico-ready > /dev/null 2>&1")
            self.execute_shell_command('kubeadm reset --cert-dir="/etc/kubernetes/pki" --ignore-preflight-errors=all --force > /dev/null 2>&1')
            self.execute_shell_command("rm -rf /etc/cni/ /etc/kubernetes/ /root/.kube/ /home/tveapi/.kube/ /var/lib/etcd/* > /dev/null 2>&1")
            # (vstojkov) NFV-18455 If file is available, don't remove disks from /var/lib/kubestore for next licensing (this is user files
            # and isn't good user experience to delete them).
            if not os.path.exists('/tmp/unregister_license'):
                self.execute_shell_command("rm -rf /var/lib/kubestore/* > /dev/null 2>&1")
        except Exception as ex:
            self.logger.error('cluster reset failed with error {}'.format(ex))
            return False

        return True

    def cluster_join_service_method(self, controller):
        join_count = 0

        # Set remote registry
        self.create_registry(controller)

        while True:
            join_count += 1
            try:
                self.logger.debug("uCPE is going to join controller {}, join retry {}".format(controller, join_count))
                token = self.get_auth_token_tveapi_controller(controller)  # get auth token
                if token is None:
                    raise Exception("Could not get tveapi token!")

                dns = os.popen("resolvectl status | grep br-data0 -A 3 | awk 'NR==4{print $4}'").read().rstrip('\n')
                self.logger.debug("uCPE join service - dns: {}".format(dns))

                token_k8s = self.get_auth_token_kubernetes(token, controller)  # get auth token
                if token_k8s is None:
                    raise Exception("Could not get kubernetes token!")
                else:
                    self.logger.debug("kubernetes cluster token: %s" % token_k8s)

                self.logger.debug("uCPE join cluster retry number %d" % join_count)
                self.execute_shell_command("kubeadm join {}:6443 --token {} --discovery-token-unsafe-skip-ca-verification > /dev/null 2>&1".format(controller, token_k8s))
                system = systemresources_api.SysrepoSystemResources(app_name="set systemresources")
                system.refresh_k8s_resources()
                #res = os.system("ifconfig | grep vpp1 >/dev/null 2>&1")
                #count = 15
                #while count >= 0:
                #    self.logger.debug("uCPE vppctl stn rule retry %d" % count)
                #    if not res:
                #        ip = os.popen('ifconfig vpp1 | grep "inet"').read().split()[1]
                #        if ip:
                #            if dns:
                #                os.system('systemd-resolve --interface vpp1 --set-dns {}'.format(dns))
                #            os.system('vppctl stn rule address {} interface tap0'.format(ip))
                #            break
                #    count -= 1
                #    sleep(10)
                #    res = os.system("ifconfig | grep vpp1 >/dev/null 2>&1")

                break
            except Exception as ex:
                self.logger.error('cluster join failed with error: {}'.format(ex))
                wait_sec = 10
                self.logger.error("Waiting {} seconds for the next try to join kubernetes cluster ...".format(wait_sec))
                time.sleep(wait_sec)
        return

    def start_cluster_join_service(self, controller=default_k8s_controller):
        self.logger.debug('start_cluster_join_service method %s' % controller)
        # Create remote registry
        self.create_registry(controller)
        #self.execute_shell_command("systemctl restart kubernetes-join@{}.service 2>/dev/null || true".format(controller))
        self.join_process = multiprocessing.Process(target=cluster_join_process, args=(self,controller,))
        self.join_process.daemon = True
        self.join_process.start()

    # not used any more
    def cluster_join(self):
        controller = self.get_controller()
        token = self.get_auth_token_tveapi_controller(controller)  # get auth token
        if token is None:
            return

        token_k8s = self.get_auth_token_kubernetes(token, controller)  # get auth token
        if token_k8s is None:
            return
        else:
            self.logger.debug("kubernetes cluster token: %s" % token_k8s)

        self.logger.info("uCPE is going to join cluster")

        dns = os.popen("resolvectl status | grep br-data0 -A 3 | awk 'NR==4{print $4}'").read().rstrip('\n')
        try:
            #subprocess.check_call(['kubeadm', 'join', '{}:6443'.format(controller), '--token', '{}'.format(token_k8s),
            # '--discovery-token-unsafe-skip-ca-verification', '> /dev/null', '2>&1'])
            self.execute_shell_command("kubeadm join {}:6443 --token {} --discovery-token-unsafe-skip-ca-verification > /dev/null 2>&1".format(controller, token_k8s))
            system = systemresources_api.SysrepoSystemResources(app_name="set systemresources")
            system.refresh_k8s_resources()
            #res = os.system("ifconfig | grep vpp1 >/dev/null 2>&1")
            #if not res:
            #    ip = os.popen('ifconfig vpp1 | grep "inet"').read().split()[1]
            #    if ip:
            #        if dns:
            #            os.system('systemd-resolve --interface vpp1 --set-dns {}'.format(dns))
            #            self.logger.debug("uCPE join service - dns configured: {}".format(dns))
            #        os.system('vppctl stn rule address {} interface tap0'.format(ip))

        except Exception as ex:
            self.logger.error('cluster join failed with error: {}'.format(ex))
        return

    def get_worker_status(self):
        controller = self.get_controller()
        token = self.get_auth_token_tveapi_controller(controller)  # get auth token
        if token is None:
            return False

        status = self.get_worker_status_from_kubernetes(token)
        if status is None:
            return False
        else:
            self.logger.error(" Worker ready status in cluster: %s" % status)
            return status

    def __bind_interface(self, interface):
        if 'data' in interface or 'tun' in interface:
            interface = 'br-' + interface

        self.write_bind_iface_in_file(interface)
        # We prioritize IPv6
        ip = self.get_interface_ip6_address(interface) or self.get_interface_ip4_address(interface)
        if ip == '':
            self.logger.error(f"Error getting IP of interface {interface} when setting it as bind interface")
            return

        with open(KUBELET_CONFIG) as f:
            file_data = f.read()
            f.seek(0)
            lines = f.readlines()
            for line in lines:
                for flag in line.split():
                    if "--node-ip" in flag:
                        node_ip = flag
                        file_data = file_data.replace(node_ip, f'--node-ip={ip}"')
                        self.logger.info(f'Changing --node-ip flag to {ip} in kubelet config')
                        break
                else:
                    continue
                break
            else:
                for line in lines:
                    if line.startswith('Environment='):
                        file_data = file_data.replace(line, line[:-2] + f' --node-ip={ip}"\n')
                        break

        with open(KUBELET_CONFIG, "w+") as f:
            f.write(file_data)

        self.execute_shell_command("systemctl daemon-reload")
        self.execute_shell_command("systemctl restart kubelet.service")

        # file_modified = False
        # with open(CALICO_VPP_CONFIG, "r") as f:
        #     list_doc = list(yaml.safe_load_all(f))
        #     if type(list_doc[0]) is list:
        #         list_doc = list_doc[0]
        #     for sense in list_doc:
        #         try:
        #             if sense['data']['vpp_dataplane_interface'] != interface:
        #                 sense['data']['vpp_dataplane_interface'] = interface
        #                 file_modified = True
        #         except KeyError:
        #             pass
        # if file_modified:
        #     self.logger.info(f'Setting {interface} as calico-vpp dataplane interface')
        #     last_idx = len(list_doc) - 1
        #     with open(CALICO_VPP_CONFIG, "w") as f:
        #         for idx, obj in enumerate(list_doc):
        #             yaml.dump(obj, f, default_flow_style=False)
        #             if idx != last_idx:
        #                 f.write('---\n')

    def write_bind_iface_in_file(self, interface):
        with open(BIND_INTERFACE_FILE, 'w+') as f:
            f.write(interface)

    def restart_nfvtime_kubernetes_init(self):
        # Remove "DONE" from file so nfvtime-kubernetes-init can recreate the cluster
        open(K8S_SINGLE_MACHINE_FILE, 'w').close()
        self.execute_shell_command("systemctl restart nfvtime-kubernetes-init.service")

    def get_interface_ip4_address(self, interface):
        if interface == "br-tun0":
            interface = "tun0"

        ip = ''
        try:
            ip = net.ifaddresses(interface)[net.AF_INET][0]['addr']
        except (ValueError, KeyError) as e:
            pass
        return ip

    def get_interface_ip6_address(self, interface):
        ip = ''
        try:
            ip = net.ifaddresses(interface)[net.AF_INET6][0]['addr']
            if ip.startswith('fe80:'):
                ip = ''
        except (ValueError, KeyError) as e:
            pass
        return ip

    def get_formatted_ip_address(self, controller):
        try:
            return f'[{controller}]' if type(ip_address(controller)) is IPv6Address else controller
        except ValueError:
            return controller

    def __interfaces_check_ippass(self):
        ippass_maps = []
        for ifname, specs in self.get_interfaces().items():
            if 'data' not in ifname and 'lte' not in ifname and 'wifi' not in ifname and 'ppp' not in ifname:
                continue
            if 'ip-passthrough' not in specs or specs['ip-passthrough'] is False or specs['ip-passthrough'] == 'false':
                continue

            # if 'data' in ifname:
            #     physname = str("physnet%s" % (int(ifname.replace("data", "")) + 1))
            # elif 'lte' in ifname:
            #     physname = str("physnet%s" % (int(ifname.replace("lte", "")) + phy_lte_offset))
            # elif 'wifi' in ifname:
            #     if 'vap' in ifname:
            #         physname = str("physnet%s" % (int(ifname.split("_")[0].replace('wifi',''))  + int(ifname.split("_")[1].replace("vap","")) + phy_wifi_offset))
            #     else:
            #         physname = str("physnet%s" % (int(ifname.replace("wifi", "")) + phy_wifi_offset))
            # elif 'ppp' in ifname:
            #     physname = str("physnet%s" % (int(ifname.replace("ppp", "")) + phy_ppp_offset))
            #
            if 'ppp' in ifname:
                sysname = ifname
            else:
                sysname = 'br-%s' % ifname

            #
            # if 'vlan' in interface and interface['vlan'] > 0 and self.get_network_mode() != 'vpp':
            #     sysname = '%s.%d' % (sysname, interface['vlan'])

            if os.path.exists('/sys/class/net/' + sysname + '/address'):
                try:
                    with open('/sys/class/net/' + sysname + '/address', 'r') as f:
                        addr = f.readline().strip('\n')
                    if addr:
                        ippass_maps.append("%s %s" % (sysname, addr))
                except Exception as e:
                    self.logger.error('Unable to get MAC address of interface {} due to failure {}'.format(sysname, e))
                    pass

        with open('/kubernetes/ippass-map-file', 'w') as f:
            for ippass_map in ippass_maps:
                self.logger.debug("Writing {} mapping in ippass-map-file".format(ippass_map))
                f.write(ippass_map + '\n')


class ConfigObjNet(ConfigObj):
    def _write_line(self, indent_string, entry, this_entry, comment):
        """Write an individual line, for the write method"""
        # NOTE: the calls to self._quote here handles non-StringType values.
        if not self.unrepr:
            val = self._decode_element(self._quote(this_entry))
        else:
            val = repr(this_entry)
        return '%s%s%s%s%s' % (indent_string,
                               self._decode_element(self._quote(entry, multiline=False)),
                               self._a_to_u('='),
                               val,
                               self._decode_element(comment))

